;******************************************************************************
;*                                                                            *
;*                            License and T&C                                 *
;*                                                                            *
;*         GR8BIT PS/2 keyboard microcontroller microcode version 1.0b        *
;*             This software is copyleft 2011 by Eugeny Brychkov              *
;*                                                                            *
;* You can use and modify its source and executable, given two conditions are *
;* met: (a) you keep this Copyleft statement in the source code, and you do   *
;* not distribute software and resulting executable on the commercial basis   *
;* and on form the purposes, not related to GR8BIT platform, without written  *
;* permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)                *
;*                                                                            *
;******************************************************************************


; Version 1.0:	(Nov 29, 2011)	- intial release
; Version 1.0a:	(Mar 02, 2012	- minor changes in assignment of the following scan codes:
;								  0x0B, 0x0D, 0x0E, 0x11, 0x15 to match compatible keyboard layout
; Version 1.0b:	(Aug 14, 2012)	- (June 18, 2012) minor fix adding small chunk of code at the beginning of
;								  key query cycle ("thecycle:") to respond with up-to-date key status (to /X0.../X7)
;								  for current scan line/row (at /YA.../YD). Issue was identified by Jan Wilmans
;								  and exhibited itself of not giving up-to-date keystroke status if system CPU code
;								  polls for row status without writing to rscan line resister (in this case /KBDIR
;								  does not get activated and MCU's interrupt routine is not invoked)
;								- added PORTC direction toggle on F12 keypress (aka keyboard controller disable), with
;								  indication by the scroll lock (LED on = output disabled)

; KEYBOARD LAYOUT
; main numeric keys, alpha keys, TAB, ESC, F1-F5, INS and DEL and CAPS are the same
; CTLR, SHIFT function the same at both sides
; STOP=F9, SEL=F10, CLS=F11


	list		p=16f1828			; list directive to define processor
	#include	<P16F1828.inc>		; processor specific variable definitions

	
    __CONFIG _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_OFF & _MCLRE_OFF & _CP_OFF & _CPD_OFF & _BOREN_OFF & _CLKOUTEN_OFF & _IESO_OFF & _FCMEN_OFF
    __CONFIG _CONFIG2, _WRT_OFF & _PLLEN_ON & _STVREN_OFF & _BORV_19 & _LVP_OFF

	;***** VARIABLE DEFINITIONS
status_tmp1	equ	0x7B			; to save STATUS within current page
fsr_temp	equ	0x7C
w_temp		EQU	0x7D			; variable used for context saving
status_temp	EQU	0x7E			; variable used for context saving
pclath_temp	EQU	0x7F			; variable used for context saving

row1		equ	0x20
row2		equ	0x21
row3		equ	0x22
row4		equ	0x23
row5		equ	0x24
row6		equ	0x25
row7		equ	0x26
row8		equ	0x27
row9		equ	0x28
row10		equ	0x29
row11		equ	0x2A
row12		equ	0x2b			; not used by GR8BIT/MSX
row13		equ	0x2c			; not used by GR8BIT/MSX
row14		equ	0x2d			; not used by GR8BIT/MSX
row15		equ	0x2f			; not used by GR8BIT/MSX

bitmask		equ	0x30
outbyte		equ	0x31
parity		equ	0x32
tmp1		equ	0x33
tmp2		equ	0x34
tmp3		equ	0x35
releaseact	equ	0x36
scanline	equ	0x37
lockkeys	equ	0x40
locktmp1	equ	0x41
locktmp2	equ	0x42
isr_tmp		equ	0x43			; interrup service routine temporary variable

	; extended key definitions
RALTKEY		equ	0x11	; E0 11 = RALT
RCTRLKEY	equ	0x14	; E0 14 = RCTRL
PADSLKEY	equ	0x4A	; E0 4A = PAD /
PADENTKEY	equ	0x5A	; E0 5A = PAD ENTER
ENDKEY		equ	0x69	; E0 69 = END
LEFTKEY		equ	0x6B	; E0 6B = LEFT
HOMEKEY		equ	0x6C	; E0 6C = HOME
INSKEY		equ	0x70	; E0 70 = INS
DELKEY		equ	0x71	; E0 71 = DEL
DOWNKEY		equ	0x72	; E0 72 = DOWN
RIGHTKEY	equ	0x74	; E0 74 = RIGHT
UPKEY		equ	0x75	; E0 75 = UP
PGDNKEY		equ	0x7A	; E0 7A = PGDN
PGUPKEY		equ	0x7D	; E0 7D = PGUP

;**********************************************************************
	ORG		0x000				; processor reset vector
  	goto		main			; go to beginning of program


	ORG		0x004				; interrupt vector location
	goto	int_handler

	; the following table should be fully within first 2K if the flash
NonETable:
	movwf	PCL
	; lookup table
	retlw	0xff		; 00 =
	retlw	0x3C		; 01 = F9				; STOP Equivalent key
	retlw	0xff		; 02 = 
	retlw	0x39		; 03 = F5
	retlw	0x37		; 04 = F3
	retlw	0x35		; 05 = F1
	retlw	0x36		; 06 = F2
	retlw	0xfe				; --------------------------------------0xff		; 07 = F12
	retlw	0xff		; 08 = 
	retlw	0x3E		; 09 = F10				; SELECT Equivalent key
	retlw	0x15		; 0A = F8
	retlw	0xff		; 0B = F6
	retlw	0x38		; 0C = F4
	retlw	0x3B		; 0D = TAB
	retlw	0x11		; 0E = ~`
	retlw	0xff		; 0F = 
	retlw	0xff		; 10 = 
	retlw	0x32		; 11 = LALT
	retlw	0x30		; 12 = LSHIFT
	retlw	0xff		; 13 = 
	retlw	0x31		; 14 = LCTRL
	retlw	0x26		; 15 = Q
	retlw	0x01		; 16 = 1!
	retlw	0xff		; 17 = 
	retlw	0xff		; 18 = 
	retlw	0xff		; 19 = 
	retlw	0x2F		; 1A = Z
	retlw	0x28		; 1B = S
	retlw	0x16		; 1C = A
	retlw	0x2c		; 1D = W
	retlw	0x02		; 1E = 2@
	retlw	0xff		; 1F = 
	retlw	0xff		; 20 = 
	retlw	0x18		; 21 = C
	retlw	0x2D		; 22 = X
	retlw	0x19		; 23 = D
	retlw	0x1A		; 24 = E
	retlw	0x04		; 25 = 4$
	retlw	0x03		; 26 = 3#
	retlw	0xff		; 27 = 
	retlw	0xff		; 28 = 
	retlw	0x40		; 29 = SPACE
	retlw	0x2B		; 2A = V
	retlw	0x1B		; 2B = F
	retlw	0x29		; 2C = T
	retlw	0x27		; 2D = R
	retlw	0x05		; 2E = 5%
	retlw	0xff		; 2F = 
	retlw	0xff		; 30 = 
	retlw	0x23		; 31 = N
	retlw	0x17		; 32 = B
	retlw	0x1D		; 33 = H
	retlw	0x1C		; 34 = G
	retlw	0x2E		; 35 = Y
	retlw	0x06		; 36 = 6^
	retlw	0xff		; 37 = 
	retlw	0xff		; 38 = 
	retlw	0xff		; 39 = 
	retlw	0x22		; 3A = M
	retlw	0x1F		; 3B = J
	retlw	0x2A		; 3C = U
	retlw	0x07		; 3D = 7&
	retlw	0x08		; 3E = 8*
	retlw	0xff		; 3F = 
	retlw	0xff		; 40 = 
	retlw	0x12		; 41 = <,		; 0x14
	retlw	0x20		; 42 = K
	retlw	0x1E		; 43 = I
	retlw	0x24		; 44 = O
	retlw	0x00		; 45 = 0)
	retlw	0x09		; 46 = 9(
	retlw	0xff		; 47 = 
	retlw	0xff		; 48 = 
	retlw	0x13		; 49 = >.		; 0x11
	retlw	0x14		; 4A = ?/		; 0x15
	retlw	0x21		; 4B = L
	retlw	0x0F		; 4C = :;
	retlw	0x25		; 4D = P
	retlw	0x0A		; 4E = -_		; 0x0B
	retlw	0xff		; 4F = 
	retlw	0xff		; 50 = 
	retlw	0xff		; 51 = 
	retlw	0x10		; 52 = "'
	retlw	0xff		; 53 = 
	retlw	0x0D		; 54 = [{
	retlw	0x0B		; 55 = +=
	retlw	0xff		; 56 = 
	retlw	0xff		; 57 = 
	retlw	0x33		; 58 = CAPS LOCK
	retlw	0x30		; 59 = RSHIFT
	retlw	0x3F		; 5A = ENTER
	retlw	0x0E		; 5B = ]}
	retlw	0xff		; 5C = 
	retlw	0x0C		; 5D = |\			0x10
	retlw	0xff		; 5E = 
	retlw	0xff		; 5F = 
	retlw	0xff		; 60 = 
	retlw	0xff		; 61 = 
	retlw	0xff		; 62 = 
	retlw	0xff		; 63 = 
	retlw	0xff		; 64 = 
	retlw	0xff		; 65 = 
	retlw	0x3D		; 66 = BACKSPACE
	retlw	0xff		; 67 = 
	retlw	0xff		; 68 = 
	retlw	0x4C		; 69 = PAD 1
	retlw	0xff		; 6A = 
	retlw	0x4F		; 6B = PAD 4
	retlw	0x52		; 6C = PAD 7
	retlw	0xff		; 6D = 
	retlw	0xff		; 6E = 
	retlw	0xff		; 6F = 
	retlw	0x4B		; 70 = PAD 0
	retlw	0x57		; 71 = PAD .
	retlw	0x4D		; 72 = PAD 2
	retlw	0x50		; 73 = PAD 5
	retlw	0x51		; 74 = PAD 6
	retlw	0x53		; 75 = PAD 8
	retlw	0x3A		; 76 = ESC
	retlw	0xff		; 77 = NUM LOCK
	retlw	0xff		; 78 = F11
	retlw	0x49		; 79 = PAD +
	retlw	0x4E		; 7A = PAD 3
	retlw	0x55		; 7B = PAD -
	retlw	0x4A		; 7C = PAD *
	retlw	0x54		; 7D = PAD 9
	retlw	0xff		; 7E = SCROLL LOCK
	retlw	0xff		; 7F = 
	retlw	0xff		; 80 =
	retlw	0xff		; 81 =
	retlw	0xff		; 82 =
	retlw	0xff		; 83 = F7
	; now defining codes invoked from extended keys routine
	retlw	0x42		; 84 = INS
	retlw	0x41		; 85 = HOME
	retlw	0x34		; 86 = CODE
	retlw	0x43		; 87 = DEL
	retlw	0xff		; 88 = not defined
	retlw	0x45		; 89 = UP
	retlw	0x46		; 8A = DOWN
	retlw	0x44		; 8B = LEFT
	retlw	0x47		; 8C = RIGHT
	; end of lookup table, maximal value of W = 0x83

SetBitTable:
	movwf	PCL
	; lookup table
	retlw	0x20		;0x01
	retlw	0x01		;0x02
	retlw	0x10		;0x04
	retlw	0x02		;0x08
	retlw	0x08		;0x10
	retlw	0x04		;0x20
	retlw	0x40
	retlw	0x80
	; end of lookup table

main:
	clrf	FSR1H
	movlw	0x8e
	movwf	FSR1L		; load pointer to the TRISC

	clrf	FSR0H			; core register - in any bank
	BANKSEL	STATUS
	clrf	STATUS			; clear IRP bit (indirect addressing)
	; initializatin of system variables

	BANKSEL	PORTA
	clrf	PORTA
	movlw	0xFF
	movwf	PORTC
	movwf	PORTB

	BANKSEL	ANSELA
	clrf	ANSELA			; PORTA digital input mode
	clrf	ANSELB			; digital
	clrf	ANSELC			; digital

	; setting IOC states
	BANKSEL	IOCAN
	clrf	IOCBP
	clrf	IOCBN			; interrupt on RB channel disabled (positive and negative)
	clrf	IOCAN
	movlw	1<<IOCAP2
	movwf	IOCAP

	BANKSEL	TRISA
	clrf	TRISC			; port C is an output (/X0../X7), bits 0..7
	movlw	0xFF
	movwf	TRISB			; port B is an input (/YA../YD), bits 4..7
	movwf	TRISA			; port A is input (including KDATA and KBDCLK)

	; clock frequency select: 32 MHz
	BANKSEL	OSCCON
	movlw	0xf0	;0xFA		;0xF0
	movwf	OSCCON			; PLL on, 8MHz

	; initialize key state registers - should be performed before enabling interrupts
	BANKSEL	0
	movlw	0xFF
	movwf	row1
	movwf	row2
	movwf	row3
	movwf	row4
	movwf	row5
	movwf	row6
	movwf	row7
	movwf	row8
	movwf	row9
	movwf	row10
	movwf	row11
	movwf	row12
	movwf	row13
	movwf	row14
	movwf	row15
	clrf	scanline

	movf	PORTA,W
	movwf	lockkeys		; save current state of the lock keys

start_up:
	; Waiting for keyboard for some time to set clock and data lines active
	clrf	tmp1
	clrf	tmp2
	movlw	0x04
	movwf	tmp3
sense_startup:
	btfsc	PORTA,1
	goto	next_scan
	btfss	PORTA,0
	goto	both_low
next_scan:
	decfsz	tmp1,1
	goto	sense_startup
	decfsz	tmp2,1
	goto	sense_startup
	decfsz	tmp3,1
	goto	sense_startup

	goto	start_up

	; code received - testing for PON complete
both_low:
	call	ReadKeyboard
	movf	outbyte,W
	sublw	0xAA				; PON sequence completed
	btfsc	STATUS,Z
	goto	PONOk				; first code received = PON Complete
	movlw	0xFF				; resetting keyboard
	call	WriteKeyboard
	goto	start_up

PONOk:
	; PON passed, now flash leds
	movlw	0x01
	movwf	releaseact
flash_leds:
	movlw	0xED
	call	WriteKeyboard
	call	ReadKeyboard		; should be 0xFA
	movf	outbyte,w
	sublw	0xFA
	btfss	STATUS,Z
	goto	flash_leds

led_val:
	movf	releaseact,W
	andlw	0x07
	call	WriteKeyboard		; write led status
	call	ReadKeyboard
	movf	outbyte,w
	sublw	0xFA
	btfss	STATUS,Z
	goto	led_val

	call	OneSecDelay
	bcf		STATUS,C
	rlf		releaseact,1
	btfss	releaseact,4		; fourth step: exit
	goto	flash_leds

	; setting maximal typematic rate
typem_rate:
	movlw	0xF3
	call	WriteKeyboard
	call	ReadKeyboard		; should be 0xFA
	movf	outbyte,w
	sublw	0xFA
	btfss	STATUS,Z
	goto	typem_rate

typem_r1:
	movlw	0x00
	call	WriteKeyboard		; 0x00 - fastest
	call	ReadKeyboard		; should be 0xFA
	movf	outbyte,w
	sublw	0xFA
	btfss	STATUS,Z
	goto	typem_r1

	; enable interrupts
	BANKSEL	INTCON
	movlw	(1<<GIE) | (1<<IOCIE)	; enable global interrupts and interrupt-on-change
	movwf	INTCON

kbd_clear:
	; Keyboard enable, clear buffers
	movlw	0xF4
	call	WriteKeyboard
	call	ReadKeyboard		; should be 0xFA
	movf	outbyte,w
	sublw	0xFA
	btfss	STATUS,Z
	goto	kbd_clear

thecycle:
; change in order to have return scan line changes
; if no /KBDIR clocking occurs (no writes to port AA,
; only continuous reads from A9 are performed)
	swapf		PORTB,W			; swap nibbles of port B and put result into W
	andlw		0x0F			; suppress bits 4..7
	addlw		row1
	movwf		FSR0L
	moviw		[FSR0]			; get value at the address pointed by FSR0
	movwf		PORTC
; the above contents actually equal to interrupt procedure handler's code

	clrf	releaseact			; not "release" action
action:
	movf	PORTA,W				; read port a input pins to process locks
	movwf	locktmp2
	xorwf	lockkeys,W
	andlw	0x30				; RA5 and RA4
	btfsc	STATUS,Z
	goto	nolockchange

	call	WriteLEDStatus

nolockchange:
	call	ReadKeyboard
	comf	outbyte,1
	btfsc	STATUS,Z
	goto	action			; 0xFF code - no action

	comf	outbyte,1
	btfsc	STATUS,Z
	goto	action			; 0x00 code - no action

	; testing for code < 0x83
	movf	outbyte,W
	addlw	0x7C				; 0x84 gives carry flag
	btfsc	STATUS,C
	goto	extendedcode

samebutextended:			; label to jump from extended code processing
	movf	outbyte,W
directtranslate:
	movwf	tmp1			; save offset
	incf	tmp1,1			; increment / correct pointer
	movlw	LOW NonETable
	addwf	tmp1,1
	movlw	HIGH NonETable
	btfsc	STATUS,C
	addlw	1
	movwf	PCLATH
	movf	tmp1,W

	call	NonETable
	movwf	outbyte

	; check for F12 key (keyboard controller enable/disable output)
	sublw	0xfe
	btfss	STATUS,Z
	goto	NoKbdDisable

	; check for F12 press action. If release, do not toggle kbd state
	btfss	releaseact,0
	comf 	INDF1,1			; invert direction of the /X lines
	call	WriteLEDStatus	; and update keyboard Num Lock indicator status
	goto	kbd_clear

NoKbdDisable:
	sublw	0x01			; testing for 0xff (previously substracted from 0xfe)
	btfsc	STATUS,Z
	goto	kbd_clear		; erroneous scan code received: clear buffer and restart

	movf	outbyte,W
	andlw	0x07				; bit number
	movwf	tmp1			; save offset
	incf	tmp1,1			; increment / correct pointer
	movlw	LOW SetBitTable
	addwf	tmp1,F
	movlw	HIGH SetBitTable
	btfsc	STATUS,C
	addlw	1
	movwf	PCLATH
	movf	tmp1,W
	call	SetBitTable

	movwf	tmp1
	btfss	releaseact,0		; bit0 = 1 => key released, no inversion, skip next instruction
	comf	tmp1,1				; invert value

	; calculate memory pointer
	rrf		outbyte,1
	rrf		outbyte,1
	rrf		outbyte,W
	andlw	0x0F					; ensure pointer < 16

	; write value to the key state table
	addlw	row1
	movwf	FSR0L
	moviw	[FSR0]
	btfss	releaseact,0		; bit0 = 1 => OR operation, otherwise AND
	goto	andop
	iorwf	tmp1,W
	goto	endlogical
andop:
	andwf	tmp1,W
endlogical:
	movwi	[FSR0]
	goto	keyfinished

extendedcode:
	; checking for extended key type 1 (0xE0)
	movf	outbyte,W
	sublw	0xE0			; release key identifier
	btfss	STATUS,Z
	goto	ext1code
extagain:					; goes here when ext key is released
	call	ReadKeyboard	; read next code
	movf	outbyte,W
	sublw	RALTKEY			; E0 11 = RALT
	btfss	STATUS,Z
	goto	afteralt
	movlw	0x86			; CODE key "scan code" for translation to 0x34 
	movwf	outbyte
	goto	samebutextended
afteralt:
	movf	outbyte,W
	sublw	RCTRLKEY		; E0 14 = RCTRL
	btfsc	STATUS,Z
	goto	samebutextended
	movf	outbyte,W
	sublw	PADSLKEY		; E0 4A = PAD /
	btfsc	STATUS,Z
	goto	samebutextended
	movf	outbyte,W
	sublw	PADENTKEY		; E0 5A = PAD ENTER
	btfsc	STATUS,Z
	goto	samebutextended
	movf	outbyte,W
	sublw	ENDKEY			; E0 69 = END
	btfsc	STATUS,Z
	goto	samebutextended
	movf	outbyte,W
	sublw	PGDNKEY			; E0 7A = PGDN
	btfsc	STATUS,Z
	goto	samebutextended
	movf	outbyte,W
	sublw	PGUPKEY			; E0 7D = PGUP
	btfsc	STATUS,Z
	goto	samebutextended	

	; now processing kays with different meanings: INS, DEL, and arrows
	movf	outbyte,W
	sublw	INSKEY			; E0 70 = INS
	btfss	STATUS,Z
	goto	delkey
	movlw	0x84
	goto	directtranslate	; translate 0x84 directly to the scan line/bit
delkey:
	movf	outbyte,W
	sublw	DELKEY			; E0 71 = DEL
	btfss	STATUS,Z
	goto	upkey
	movlw	0x87
	goto	directtranslate	; translate 0x87 directly to the scan line/bit
upkey:
	movf	outbyte,W
	sublw	UPKEY			; E0 75 = UP
	btfss	STATUS,Z
	goto	downkey
	movlw	0x89
	goto	directtranslate	; translate 0x87 directly to the scan line/bit
downkey:
	movf	outbyte,W
	sublw	DOWNKEY			; E0 72 = DOWN
	btfss	STATUS,Z
	goto	leftkey
	movlw	0x8A
	goto	directtranslate	; translate 0x8A directly to the scan line/bit
leftkey:
	movf	outbyte,W
	sublw	LEFTKEY			; E0 6B = LEFT
	btfss	STATUS,Z
	goto	rightkey
	movlw	0x8B
	goto	directtranslate	; translate 0x8B directly to the scan line/bit
rightkey:
	movf	outbyte,W
	sublw	RIGHTKEY		; E0 74 = RIGHT
	btfss	STATUS,Z
	goto	homekey
	movlw	0x8C
	goto	directtranslate	; translate 0x8C directly to the scan line/bit
homekey:
	movf	outbyte,W
	sublw	HOMEKEY			; E0 6C = HOME
	btfss	STATUS,Z
	goto	releaseextended
	movlw	0x85
	goto	directtranslate	; translate 0x8C directly to the scan line/bit

releaseextended:
	movf	outbyte,W
	sublw	0xF0			; E0 F0 * = released extended key
	btfss	STATUS,Z
	goto	kbd_clear		; reset keyboard as invalid value was received

	comf	releaseact,1
	goto	extagain		; turned release key mode on and goes to translation

ext1code:
	; here we check for only one key: E1 14 77 E1 F0 14 F0 77 = PAUSE/BREAK
	movf	outbyte,W
	sublw	0xE1			; release key identifier
	btfsc	STATUS,Z
	goto	kbd_clear		; clear break key press

	movf	outbyte,W
	sublw	0xEE			; POST finished ID
	btfsc	STATUS,Z
	goto	kbd_clear
	movf	outbyte,W
	sublw	0x00			; POST finished ID
	btfsc	STATUS,Z
	goto	kbd_clear
	movf	outbyte,W
	sublw	0xFF			; POST finished ID
	btfsc	STATUS,Z
	goto	kbd_clear
	movf	outbyte,W
	sublw	0xFE			; POST finished ID
	btfsc	STATUS,Z
	goto	kbd_clear

	; checking for release key identifier (0xF0)
	movf	outbyte,W
	sublw	0xF0			; release key identifier
	btfss	STATUS,Z
	goto	keyfinished

	; key released: change mode of the release action
	comf	releaseact,1
	goto	action

keyfinished:
	goto	thecycle


; Read keyboard byte, RA0/RA1 lines should be input. Output = outbyte
ReadKeyboard:
	; set bit mask for output byte, clear output byte
	movlw	0x01
	movwf	bitmask
	clrf	outbyte
	clrf	parity

	; set CLK/DATA as inputs
	BANKSEL	TRISA
	movlw	0xFF			; RA1/RA0 as inputs
	movwf	TRISA			; port A change mode
	BANKSEL	0

	; wait for RA1 (CLK) to become active
readkbd_1:
	btfsc	PORTA,1
	goto	readkbd_1

	; if CLK is active and DATA is inactive, wait for data to activate or CLK deactivate
	btfsc	PORTA,0
	goto	readkbd_1

	; we come here when CLK and DATA lines are active - starting receiving data
readkbd_2:
	call	WaitForNegEdge
	; KDATA is in PORTA, bit 0 (RA0)
	movf	bitmask,W		; read current bit mask into W
	btfss	PORTA,0			; if data bit is 1, then execute next instruction
	goto	read_zero
	iorwf	outbyte,1		; OR outbyte with W (current mask) and store into outbyte
	incf	parity,1
read_zero:
	bcf		STATUS,C		; clear carry before it is inserted into the bit 0 of bitmask
	rlf		bitmask,1		; rotate bitmask left, place result into bitmask
	btfss	STATUS,C		; masks's bit is shifted to C flag - all 8 bits were read
	goto	readkbd_2

	; process parity
	incf	parity,1
	call	WaitForNegEdge
	movf	PORTA,W
	xorwf	parity,1
;	btfsc	parity,0		; if parity bit = 1 after xor, then parity mismatch
;	goto	recv_error

	; process stop bit
	call	WaitForNegEdge
;	btfss	PORTA,0
;	goto	recv_error
;	call	WaitForPosEdge

CTSandReturn:
	movlw	0xFD			; // RA1 is output, RA0 is input
	movwf	PORTA			; bit 1 is zero
	; set CLK as output and send clear-to-send (CLK=active)
	BANKSEL	TRISA
	movwf	TRISA			; port A change mode
	BANKSEL	0
	return

recv_error:
	movlw	0xFE			; resend command
	movwf	outbyte
	call	WriteKeyboard
	goto	ReadKeyboard	; Keyboard

; ------------------- Write to the keyboard --------------------
;	in: W
; --------------------------------------------------------------
WriteKeyboard:
	movwf	outbyte
	; set bit mask for output byte, set initial parity value
	movlw	0x01
	movwf	bitmask
	movwf	parity			; to be equal to 0x1

	; output RA1 (CLK) low, and RA0 (DATA) high
	BANKSEL	PORTA
	movlw	0xFD
	movwf	PORTA

	; set bank 1 to change direction of RA1/RA0
	BANKSEL	TRISA
	movwf	TRISA			; port A change mode: RA1 as output
	BANKSEL	0

	; now wait for 72 microseconds: 31.25ns*3*256*3
	movlw	0x05
	movwf	tmp2
	clrf	tmp1
	; three cycles per two following commands, 125ns for single cycle
	; equals to 96 microseconds
writedelay:
	decfsz	tmp1,1
	goto	writedelay
	decfsz	tmp2,1
	goto	writedelay

	; set DATA low and release CLK
	movlw	0xFE			; RA0 as output
	movwf	PORTA
	; set bank 1 to change direction of RA1/RA0
	BANKSEL	TRISA
	movwf	TRISA			; port A change mode
	BANKSEL	0

writekbd_1:
	call	WaitForPosEdge

	rrf		outbyte,1
	movlw	0xFE
	btfss	STATUS,C		; if C is set, then execute next instruction
	goto	zerobit	
	incf	parity,1
	movlw	0xFF
zerobit:
	movwf	PORTA			; out bit to the RA0 (KDATA)

	; checking for if all 8 bits were processed
	rlf		bitmask,1
	btfss	STATUS,C
	goto	writekbd_1

	call	WaitForPosEdge	; wait for parity bit cycle
	; writing odd parity
	movf	parity,W
	iorlw	0xFE			; setting all bits except bit 0 to 1, bit 0 contains odd parity bit
	movwf	PORTA

	call	WaitForPosEdge	; wait for release cycle
	; releasing clock and data lines
	; set bank 1 to change direction of RA1/RA0
	movlw	0xFF
	movwf	PORTA

	BANKSEL	TRISA
	movwf	TRISA			; port A change mode: RA1/RA0 as outputs
	BANKSEL	0

	call	WaitForPosEdge	; wait for acknowledge cycle
	btfsc	PORTA,0
	goto	write_error		; in case ack bit is 0

	; wait for CLK line to be released
writekbd_2:
	btfss	PORTA,1
	goto	writekbd_2

	goto	CTSandReturn

write_error:
	movlw	0xFF			; reset
	call	WriteKeyboard
	return

; ---------------- Wait for CLK negative edge ------------------
;	in: none
;	out: none
;	corrupt: none
; --------------------------------------------------------------
WaitForNegEdge:
	; wait while CLK is low
TrackPos:
	btfss	PORTA,1
	goto	TrackPos

	; wait while CLK is high
TrackNeg:
	btfsc	PORTA,1
	goto	TrackNeg
	return

; ---------------- Wait for CLK positive edge ------------------
;	in: none
;	out: none
;	corrupt: none
; --------------------------------------------------------------
WaitForPosEdge:
	; wait while CLK is high
TrackNeg1:
	btfsc	PORTA,1
	goto	TrackNeg1

	; wait while CLK is low
TrackPos1:
	btfss	PORTA,1
	goto	TrackPos1
	return

OneSecDelay:
	clrw
	movwf	tmp1
	movwf	tmp2
	movlw	0x10		; 0x40
	movwf	tmp3
	; approximate delay at 8MHz: 3*256*125ns*256*32+3*256*125ns*32+3*32*125ns = 0.78 seconds
delayloop:
	decfsz	tmp1,f
	goto	delayloop
	decfsz	tmp2,f
	goto	delayloop
	decfsz	tmp3,f
	goto	delayloop
	return

		; E0 12 E0 7C = PRINT SCREEN
		; E1 14 77 E1 F0 14 F0 77 = PAUSE/BREAK

; set the keyboard LED status (according to Code, Caps and state of the keyboard)
WriteLEDStatus:
	movf	PORTA,W				; read port a input pins to process locks
	movwf	locktmp2
	clrf	locktmp1
	btfss	locktmp2,4
	bsf		locktmp1,2
	btfss	locktmp2,5
	bsf		locktmp1,0
	btfsc	INDF1,0				; port C is in input? (kbd disabled?)
	bsf		locktmp1,1			; turn num lock LED on
led_cmd_1:
	movlw	0xED
	call	WriteKeyboard
	call	ReadKeyboard		; should be 0xFA
	movf	outbyte,w
	sublw	0xFA
	btfss	STATUS,Z
	goto	led_cmd_1

led_cmd_2:
	movf	locktmp1,W
	call	WriteKeyboard		; write led status
	call	ReadKeyboard
	movf	outbyte,w
	sublw	0xFA
	btfss	STATUS,Z
	goto	led_cmd_2

	movf	locktmp2,W
	movwf	lockkeys
	return

int_handler:
	; context is automatically saved
	BANKSEL		PORTA
	swapf		PORTB,W			; swap nibbles of port B and put result into W
	; important: input pins 3..0 of PORTB read as 0s!
	;andlw		0x0F			; suppress bits 4..7
	addlw		row1
	movwf		FSR0L
	moviw		[FSR0]			; get value at the address pointed by FSR0
	movwf		PORTC
	BANKSEL		IOCAF
	clrf		IOCAF
	retfie						; return from interrupt

	END                       ; directive 'end of program'